import { ref, reactive, computed } from 'vue';

export interface Sprite {
  img: HTMLImageElement;
  width: number;
  height: number;
  x: number;
  y: number;
  name: string;
  id: string;
  uploadOrder: number;
}

export interface CellSize {
  width: number;
  height: number;
}

export interface AnimationState {
  canvas: HTMLCanvasElement | null;
  ctx: CanvasRenderingContext2D | null;
  currentFrame: number;
  isPlaying: boolean;
  frameRate: number;
  lastFrameTime: number;
  animationId: number | null;
  slider: HTMLInputElement | null;
  manualUpdate: boolean;
}

export function useSpritesheetStore() {
  const sprites = ref<Sprite[]>([]);
  const canvas = ref<HTMLCanvasElement | null>(null);
  const ctx = ref<CanvasRenderingContext2D | null>(null);
  const cellSize = reactive<CellSize>({ width: 0, height: 0 });
  const columns = ref(4); // Default number of columns
  const draggedSprite = ref<Sprite | null>(null);
  const dragOffset = reactive({ x: 0, y: 0 });
  const isShiftPressed = ref(false);
  const isModalOpen = ref(false);

  const animation = reactive<AnimationState>({
    canvas: null,
    ctx: null,
    currentFrame: 0,
    isPlaying: false,
    frameRate: 10,
    lastFrameTime: 0,
    animationId: null,
    slider: null,
    manualUpdate: false
  });

  const notification = reactive({
    isVisible: false,
    message: '',
    type: 'success' as 'success' | 'error'
  });

  function addSprites(newSprites: Sprite[]) {
    sprites.value.push(...newSprites);
    sprites.value.sort((a, b) => a.uploadOrder - b.uploadOrder);
    updateCellSize();
    autoArrangeSprites();
  }

  function updateCellSize() {
    if (sprites.value.length === 0) return;

    let maxWidth = 0;
    let maxHeight = 0;

    sprites.value.forEach(sprite => {
      maxWidth = Math.max(maxWidth, sprite.width);
      maxHeight = Math.max(maxHeight, sprite.height);
    });

    cellSize.width = maxWidth;
    cellSize.height = maxHeight;

    updateCanvasSize();
  }

  function updateCanvasSize() {
    if (!canvas.value || sprites.value.length === 0) return;

    const totalSprites = sprites.value.length;
    const cols = columns.value;
    const rows = Math.ceil(totalSprites / cols);

    canvas.value.width = cols * cellSize.width;
    canvas.value.height = rows * cellSize.height;
  }

  function autoArrangeSprites() {
    if (sprites.value.length === 0) return;

    sprites.value.forEach((sprite, index) => {
      const column = index % columns.value;
      const row = Math.floor(index / columns.value);

      sprite.x = column * cellSize.width;
      sprite.y = row * cellSize.height;
    });

    renderSpritesheetPreview();

    if (!animation.isPlaying && animation.manualUpdate && isModalOpen.value) {
      renderAnimationFrame(animation.currentFrame);
    }
  }

  function renderSpritesheetPreview(showGrid = true) {
    if (!ctx.value || !canvas.value || sprites.value.length === 0) return;

    ctx.value.clearRect(0, 0, canvas.value.width, canvas.value.height);

    if (showGrid) {
      drawGrid();
    }

    sprites.value.forEach(sprite => {
      ctx.value!.drawImage(
        sprite.img,
        sprite.x,
        sprite.y
      );
    });
  }

  function drawGrid() {
    if (!ctx.value || !canvas.value) return;

    ctx.value.strokeStyle = '#333';
    ctx.value.lineWidth = 1;

    // Draw vertical lines
    for (let x = 0; x <= canvas.value.width; x += cellSize.width) {
      ctx.value.beginPath();
      ctx.value.moveTo(x, 0);
      ctx.value.lineTo(x, canvas.value.height);
      ctx.value.stroke();
    }

    // Draw horizontal lines
    for (let y = 0; y <= canvas.value.height; y += cellSize.height) {
      ctx.value.beginPath();
      ctx.value.moveTo(0, y);
      ctx.value.lineTo(canvas.value.width, y);
      ctx.value.stroke();
    }
  }

  function highlightSprite(spriteId: string) {
    if (!ctx.value || !canvas.value) return;

    const sprite = sprites.value.find(s => s.id === spriteId);
    if (!sprite) return;

    // Calculate the cell coordinates
    const cellX = Math.floor(sprite.x / cellSize.width);
    const cellY = Math.floor(sprite.y / cellSize.height);

    // Briefly flash the cell
    ctx.value.save();
    ctx.value.fillStyle = 'rgba(0, 150, 255, 0.3)';
    ctx.value.fillRect(
      cellX * cellSize.width,
      cellY * cellSize.height,
      cellSize.width,
      cellSize.height
    );
    ctx.value.restore();

    // Reset after a short delay
    setTimeout(() => {
      renderSpritesheetPreview();
    }, 500);
  }

  function clearAllSprites() {
    if (!canvas.value || !ctx.value) return;

    sprites.value = [];
    canvas.value.width = 400;
    canvas.value.height = 300;
    ctx.value.clearRect(0, 0, canvas.value.width, canvas.value.height);

    if (animation.canvas && animation.ctx) {
      animation.canvas.width = 200;
      animation.canvas.height = 200;
      animation.ctx.clearRect(0, 0, animation.canvas.width, animation.canvas.height);
    }

    animation.currentFrame = 0;
    isModalOpen.value = false;
  }

  function downloadSpritesheet() {
    if (sprites.value.length === 0 || !canvas.value) {
      showNotification('No sprites to download', 'error');
      return;
    }

    const tempCanvas = document.createElement('canvas');
    tempCanvas.width = canvas.value.width;
    tempCanvas.height = canvas.value.height;
    const tempCtx = tempCanvas.getContext('2d');

    if (!tempCtx) {
      showNotification('Failed to create download context', 'error');
      return;
    }

    tempCtx.clearRect(0, 0, tempCanvas.width, tempCanvas.height);

    sprites.value.forEach(sprite => {
      tempCtx.drawImage(
        sprite.img,
        sprite.x,
        sprite.y
      );
    });

    const link = document.createElement('a');
    link.download = 'noxious-spritesheet.png';
    link.href = tempCanvas.toDataURL('image/png');
    document.body.appendChild(link);
    link.click();
    document.body.removeChild(link);

    showNotification('Spritesheet downloaded successfully');
  }

  function startAnimation() {
    if (sprites.value.length === 0 || !animation.canvas) return;

    animation.isPlaying = true;
    animation.lastFrameTime = performance.now();
    animation.manualUpdate = false;

    animation.canvas.width = cellSize.width;
    animation.canvas.height = cellSize.height;

    animationLoop();
  }

  function stopAnimation() {
    animation.isPlaying = false;

    if (animation.animationId) {
      cancelAnimationFrame(animation.animationId);
      animation.animationId = null;
    }
  }

  function renderAnimationFrame(frameIndex: number) {
    if (sprites.value.length === 0 || !animation.canvas || !animation.ctx) return;

    if (animation.canvas.width !== cellSize.width ||
      animation.canvas.height !== cellSize.height) {
      animation.canvas.width = cellSize.width;
      animation.canvas.height = cellSize.height;
    }

    animation.ctx.clearRect(0, 0, animation.canvas.width, animation.canvas.height);

    const currentSprite = sprites.value[frameIndex % sprites.value.length];

    const cellX = Math.floor(currentSprite.x / cellSize.width);
    const cellY = Math.floor(currentSprite.y / cellSize.height);

    const offsetX = currentSprite.x - (cellX * cellSize.width);
    const offsetY = currentSprite.y - (cellY * cellSize.height);

    animation.ctx.drawImage(
      currentSprite.img,
      offsetX,
      offsetY
    );
  }

  function animationLoop(timestamp?: number) {
    if (!animation.isPlaying) return;

    const currentTime = timestamp || performance.now();
    const elapsed = currentTime - animation.lastFrameTime;
    const frameInterval = 1000 / animation.frameRate;

    if (elapsed >= frameInterval) {
      animation.lastFrameTime = currentTime;

      if (sprites.value.length > 0) {
        renderAnimationFrame(animation.currentFrame);

        animation.currentFrame = (animation.currentFrame + 1) % sprites.value.length;

        if (animation.slider) {
          animation.slider.value = animation.currentFrame.toString();
        }
      }
    }

    animation.animationId = requestAnimationFrame(animationLoop);
  }

  function showNotification(message: string, type: 'success' | 'error' = 'success') {
    notification.message = message;
    notification.type = type;
    notification.isVisible = true;

    setTimeout(() => {
      notification.isVisible = false;
    }, 3000);
  }

  return {
    sprites,
    canvas,
    ctx,
    cellSize,
    columns,
    draggedSprite,
    dragOffset,
    isShiftPressed,
    isModalOpen,
    animation,
    notification,
    addSprites,
    updateCellSize,
    updateCanvasSize,
    autoArrangeSprites,
    renderSpritesheetPreview,
    drawGrid,
    highlightSprite,
    clearAllSprites,
    downloadSpritesheet,
    startAnimation,
    stopAnimation,
    renderAnimationFrame,
    showNotification
  };
}